﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class SellEconomicValue : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertItemClassInit(gameModule, modModule);
        InsertCreateItemsInit(gameModule, modModule);
        InsertBlockInit(gameModule, modModule);
        InsertCreateBlocksInit(gameModule, modModule);
        InsertAutoCalcSellEcoVal(gameModule, modModule);
        FixGetSellPrice(gameModule, modModule);
        return true;
    }
    public void FixGetSellPrice(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "XUiM_Trader");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GetSellPrice");

        TypeDefinition blockDef = gameModule.Types.First(d => d.Name == "Block");
        FieldDefinition blockEcoVal = blockDef.Fields.First(d => d.Name == "SellEconomicValue");
        TypeDefinition itemClassDef = gameModule.Types.First(d => d.Name == "ItemClass");
        FieldDefinition itemEcoVal = itemClassDef.Fields.First(d => d.Name == "SellEconomicValue");

        Instruction updatePosition1 = null;
        Instruction updatePosition2 = null;
        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single Block::EconomicValue")
            {
                updatePosition1 = inst;
            }
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single ItemClass::EconomicValue")
            {
                updatePosition2 = inst;
            }
        }
        updatePosition1.Operand = blockEcoVal;
        updatePosition2.Operand = itemEcoVal;
    }

    public void InsertAutoCalcSellEcoVal(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "WorldStaticData");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Init" && d.Parameters.Count == 2);
        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Single ItemClass::AutoCalcEcoVal()")
            {
                insertPosition = inst.Next.Next;
            }
        }
        TypeDefinition sellEconomicValueMethod = modModule.Types.First(d => d.Name == "SellEconomicValueMethod");
        MethodDefinition autoCalcSellEcoVal = sellEconomicValueMethod.Methods.First(d => d.Name == "AutoCalcSellEcoVal");
        MethodReference autoCalcSellEcoValRef = gameModule.Import(autoCalcSellEcoVal);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_0);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, autoCalcSellEcoValRef));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Pop));
    }
    public void InsertBlockInit(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Block");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Init");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldflda && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single Block::EconomicValue")
            {
                insertPosition = inst.Next.Next.Next.Next.Next;
            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        TypeDefinition sellEconomicValueMethod = modModule.Types.First(d => d.Name == "SellEconomicValueMethod");
        MethodDefinition blockInit = sellEconomicValueMethod.Methods.First(d => d.Name == "BlockInit");
        MethodReference blockInitRef = gameModule.Import(blockInit);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, blockInitRef));


        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }


    }

    public void InsertItemClassInit(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "ItemClass");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Init");

        Instruction insertPosition = null;
        foreach(Instruction inst in methodDef.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Ldflda && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single ItemClass::EconomicValue")
            {
                insertPosition = inst.Next.Next.Next.Next.Next;
            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        TypeDefinition sellEconomicValueMethod = modModule.Types.First(d => d.Name == "SellEconomicValueMethod");
        MethodDefinition itemClassInit = sellEconomicValueMethod.Methods.First(d => d.Name == "ItemClassInit");
        MethodReference itemClassInitRef = gameModule.Import(itemClassInit);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, itemClassInitRef));


        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }


    }
    public void InsertCreateBlocksInit(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "BlocksFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "CreateBlocks" && d.Parameters.Count == 2);

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single Block::EconomicValue")
            {
                insertPosition = inst.Next;
            }
        }

        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }


        TypeDefinition sellEconomicValueMethod = modModule.Types.First(d => d.Name == "SellEconomicValueMethod");
        MethodDefinition createBlocks = sellEconomicValueMethod.Methods.First(d => d.Name == "CreateBlocks");
        MethodReference createBlocksRef = gameModule.Import(createBlocks);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[16]);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, createBlocksRef));


        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }


    }

    
    public void InsertCreateItemsInit(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "ItemClassesFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "CreateItems");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single ItemClass::EconomicValue")
            {
                insertPosition = inst.Next;
            }
        }

        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }


        TypeDefinition sellEconomicValueMethod = modModule.Types.First(d => d.Name == "SellEconomicValueMethod");
        MethodDefinition createItems = sellEconomicValueMethod.Methods.First(d => d.Name == "CreateItems");
        MethodReference createItemsRef = gameModule.Import(createItems);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[21]);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[11]));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, createItemsRef));


        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }


    }


    public void AddItemClassField(ModuleDefinition module)
    {
        TypeReference tr = module.Import(typeof(float));
        TypeDefinition prefab = module.Types.First(d => d.Name == "ItemClass");

        AddFieldDefinition(prefab, "SellEconomicValue", FieldAttributes.Public, tr);
    }
    public void AddBlockField(ModuleDefinition module)
    {
        TypeReference tr = module.Import(typeof(float));
        TypeDefinition prefab = module.Types.First(d => d.Name == "Block");

        AddFieldDefinition(prefab, "SellEconomicValue", FieldAttributes.Public, tr);
    }
    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==SellEconomicValue Patcher===");
        AddItemClassField(module);
        AddBlockField(module);
        return true;
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    private void SetFieldToPublic(FieldDefinition field)
    {
        field.IsFamily = false;
        field.IsPrivate = false;
        field.IsPublic = true;

    }

}
